from .constants import ALL
from .tasks import Task
from .utils import import_attr


class Gator(object):
    def __init__(
        self, conn_string, queue_name=ALL, task_class=Task, backend_class=None
    ):
        """
        A coordination for scheduling & processing tasks.

        Handles creating tasks (with options), using the backend to place tasks
        in the queue & pulling/processing tasks off the queue.

        Ex::

            from alligator import Gator

            def add(a, b):
                return a + b

            gator = Gator('locmem://')

            gator.task(add, 3, 7)

        :param conn_string: A DSN for connecting to the queue. Passed along
            to the backend.
        :type conn_string: string

        :param queue_name: (Optional) The name of the queue the tasks should be
            placed in. Defaults to ``ALL``.
        :type queue_name: string

        :param task_class: (Optional) The class to use for instantiating tasks.
            Defaults to ``Task``.
        :type task_class: class

        :param backend_class: (Optional) The class to use for instantiating
            the backend. Defaults to ``None`` (DSN detection).
        :type backend_class: class
        """
        self.conn_string = conn_string
        self.queue_name = queue_name
        self.task_class = task_class
        self.backend_class = backend_class

        if not backend_class:
            self.backend = self.build_backend(self.conn_string)
        else:
            self.backend = backend_class(self.conn_string)

    def build_backend(self, conn_string):
        """
        Given a DSN, returns an instantiated backend class.

        Ex::

            backend = gator.build_backend('locmem://')
            # ...or...
            backend = gator.build_backend('redis://127.0.0.1:6379/0')

        :param conn_string: A DSN for connecting to the queue. Passed along
            to the backend.
        :type conn_string: string

        :returns: A backend ``Client`` instance
        """
        backend_name, _ = conn_string.split(":", 1)
        backend_path = "alligator.backends.{}_backend".format(backend_name)
        client_class = import_attr(backend_path, "Client")
        return client_class(conn_string)

    def len(self):
        """
        Returns the number of remaining queued tasks.

        :returns: An integer count
        """
        return self.backend.len(self.queue_name)

    def push(self, task, func, *args, **kwargs):
        """
        Pushes a configured task onto the queue.

        Typically, you'll favor using the ``Gator.task`` method or
        ``Gator.options`` context manager for creating a task. Call this
        only if you have specific needs or know what you're doing.

        If the ``Task`` has the ``is_async = False`` option, the task will be
        run immediately (in-process). This is useful for development and
        in testing.

        Ex::

            task = Task(is_async=False, retries=3)
            finished = gator.push(task, increment, incr_by=2)

        :param task: A mostly-configured task
        :type task: A ``Task`` instance

        :param func: The callable with business logic to execute
        :type func: callable

        :param args: Positional arguments to pass to the callable task
        :type args: list

        :param kwargs: Keyword arguments to pass to the callable task
        :type kwargs: dict

        :returns: The ``Task`` instance
        """
        task.to_call(func, *args, **kwargs)
        data = task.serialize()

        if task.is_async:
            task.task_id = self.backend.push(
                self.queue_name, task.task_id, data
            )
        else:
            self.execute(task)

        return task

    def pop(self):
        """
        Pops a task off the front of the queue & runs it.

        Typically, you'll favor using a ``Worker`` to handle processing the
        queue (to constantly consume). However, if you need to custom-process
        the queue in-order, this method is useful.

        Ex::

            # Tasks were previously added, maybe by a different process or
            # machine...
            finished_topmost_task = gator.pop()

        :returns: The completed ``Task`` instance
        """
        data = self.backend.pop(self.queue_name)

        if data:
            task = self.task_class.deserialize(data)
            return self.execute(task)

    def get(self, task_id):
        """
        Gets a specific task, by ``task_id`` off the queue & runs it.

        Using this is not as performant (because it has to search the queue),
        but can be useful if you need to specifically handle a task
        *right now*.

        Ex::

            # Tasks were previously added, maybe by a different process or
            # machine...
            finished_task = gator.get('a-specific-uuid-here')

        :param task_id: The identifier of the task to process
        :type task_id: string

        :returns: The completed ``Task`` instance
        """
        data = self.backend.get(self.queue_name, task_id)

        if data:
            task = self.task_class.deserialize(data)
            return self.execute(task)

    def cancel(self, task_id):
        """
        Takes an existing task & cancels it before it is processed.

        Returns the canceled task, as that could be useful in creating a new
        task.

        Ex::

            task = gator.task(add, 18, 9)

            # Whoops, didn't mean to do that.
            gator.cancel(task.task_id)

        :param task_id: The identifier of the task to process
        :type task_id: string

        :returns: The canceled ``Task`` instance
        """
        data = self.backend.get(self.queue_name, task_id)

        if data:
            task = self.task_class.deserialize(data)
            task.to_canceled()
            return task

    def execute(self, task):
        """
        Given a task instance, this runs it.

        This includes handling retries & re-raising exceptions.

        Ex::

            task = Task(is_async=False, retries=5)
            task.to_call(add, 101, 35)
            finished_task = gator.execute(task)

        :param task_id: The identifier of the task to process
        :type task_id: string

        :returns: The completed ``Task`` instance
        """
        try:
            return task.run()
        except Exception:
            if task.retries > 0:
                task.retries -= 1
                task.to_retrying()

                if task.is_async:
                    # Place it back on the queue.
                    data = task.serialize()
                    task.task_id = self.backend.push(
                        self.queue_name, task.task_id, data
                    )
                else:
                    return self.execute(task)
            else:
                raise

    def task(self, func, *args, **kwargs):
        """
        Pushes a task onto the queue.

        This will instantiate a ``Gator.task_class`` instance, configure
        the callable & its arguments, then push it onto the queue.

        You'll typically want to use either this method or the
        ``Gator.options`` context manager (if you need to configure the
        ``Task`` arguments, such as retries, is_async, task_id, etc.)

        Ex::

            on_queue = gator.task(increment, incr_by=2)

        :param func: The callable with business logic to execute
        :type func: callable

        :param args: Positional arguments to pass to the callable task
        :type args: list

        :param kwargs: Keyword arguments to pass to the callable task
        :type kwargs: dict

        :returns: The ``Task`` instance
        """
        task = self.task_class()
        return self.push(task, func, *args, **kwargs)

    def options(self, **kwargs):
        """
        Allows specifying advanced ``Task`` options to control how the task
        runs.

        This returns a context manager which will create ``Task`` instances
        with the supplied options. See ``Task.__init__`` for the available
        arguments.

        Ex::

            def party_time(task, result):
                # Throw a party in honor of this task completing.
                # ...

            with gator.options(retries=2, on_success=party_time) as opts:
                opts.task(increment, incr_by=2678)

        :param kwargs: Keyword arguments to control the task execution
        :type kwargs: dict

        :returns: An ``Options`` context manager instance
        """
        return Options(self, **kwargs)


class Options(object):
    def __init__(self, gator, **kwargs):
        """
        A context manager for specifying task execution options.

        Typically, you'd use ``Gator.options``, which creates this context
        manager for you. You probably don't want to directly use this.

        :param gator: A configured ``Gator`` instance.
        :type gator: ``Gator`` instance

        :param **kwargs: Keyword arguments to control the task execution
        :type **kwargs: dict
        """
        self.gator = gator
        self.task_kwargs = kwargs

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def task(self, func, *args, **kwargs):
        """
        Pushes a task onto the queue (with the specified options).

        This will instantiate a ``Gator.task_class`` instance, configure the
        task execution options, configure the callable & its arguments, then
        push it onto the queue.

        You'll typically call this method when specifying advanced options.

        :param func: The callable with business logic to execute
        :type func: callable

        :param args: Positional arguments to pass to the callable task
        :type args: list

        :param kwargs: Keyword arguments to pass to the callable task
        :type kwargs: dict

        :returns: The ``Task`` instance
        """
        task = self.gator.task_class(**self.task_kwargs)
        return self.gator.push(task, func, *args, **kwargs)



class Client(object):
    queues = {}
    task_data = {}

    def __init__(self, conn_string):
        """
        An in-memory ``Client``. Useful for development & testing.

        Likely not particularly thread-safe.

        :param conn_string: The DSN. Ignored.
        :type conn_string: string
        """
        # We ignore the conn_string, since everything is happening in-memory.
        pass

    def len(self, queue_name):
        """
        Returns the length of the queue.

        :param queue_name: The name of the queue. Usually handled by the
            ``Gator`` instance.
        :type queue_name: string

        :returns: The length of the queue
        :rtype: integer
        """
        return len(self.__class__.queues.get(queue_name, []))

    def drop_all(self, queue_name):
        """
        Drops all the task in the queue.

        :param queue_name: The name of the queue. Usually handled by the
            ``Gator`` instance.
        :type queue_name: string
        """
        cls = self.__class__

        for task_id in cls.queues.get(queue_name, []):
            cls.task_data.pop(task_id, None)

        cls.queues[queue_name] = []

    def push(self, queue_name, task_id, data):
        """
        Pushes a task onto the queue.

        :param queue_name: The name of the queue. Usually handled by the
            ``Gator`` instance.
        :type queue_name: string

        :param task_id: The identifier of the task.
        :type task_id: string

        :param data: The relevant data for the task.
        :type data: string
        """
        cls = self.__class__
        cls.queues.setdefault(queue_name, [])
        cls.queues[queue_name].append(task_id)
        cls.task_data[task_id] = data
        return task_id

    def pop(self, queue_name):
        """
        Pops a task off the queue.

        :param queue_name: The name of the queue. Usually handled by the
            ``Gator`` instance.
        :type queue_name: string

        :returns: The data for the task.
        :rtype: string
        """
        cls = self.__class__
        queue = cls.queues.get(queue_name, [])

        if queue:
            task_id = queue.pop(0)
            return cls.task_data.pop(task_id, None)

    def get(self, queue_name, task_id):
        """
        Pops a specific task off the queue by identifier.

        :param queue_name: The name of the queue. Usually handled by the
            ``Gator`` instance.
        :type queue_name: string

        :param task_id: The identifier of the task.
        :type task_id: string

        :returns: The data for the task.
        :rtype: string
        """
        # This method is *very* non-thread-safe.
        cls = self.__class__
        queue = cls.queues.get(queue_name, [])

        if queue:
            try:
                offset = queue.index(task_id)
            except ValueError:
                return None

            queue.pop(offset)
            return cls.task_data.pop(task_id, None)



from urllib.parse import urlparse

import boto3
from botocore.config import Config


class Client(object):
    def __init__(self, conn_string):
        """
        A Amazon SQS-based ``Client``.

        :param conn_string: The DSN. The region is parsed out of it.
            Should be of the format ``sqs://region/``
        :type conn_string: string
        """
        self.conn_string = conn_string
        bits = urlparse(self.conn_string)
        self.conn = self.get_connection(region=bits.hostname)
        self._queue = None

    def get_connection(self, region):
        """
        Returns a ``SQSConnection`` connection instance.
        """
        config = Config(region_name=region)
        return boto3.resource("sqs", config=config)

    def _get_queue(self, queue_name):
        if self._queue is None:
            self._queue = self.conn.get_queue_by_name(QueueName=queue_name)

        return self._queue

    def len(self, queue_name):
        """
        Returns the length of the queue.

        :param queue_name: The name of the queue. Usually handled by the
            ``Gator`` instance.
        :type queue_name: string

        :returns: The length of the queue
        :rtype: integer
        """
        queue = self._get_queue(queue_name)
        queue.load()
        return int(queue.attributes.get("ApproximateNumberOfMessages", 0))

    def drop_all(self, queue_name):
        """
        Drops all the task in the queue.

        :param queue_name: The name of the queue. Usually handled by the
            ``Gator`` instance.
        :type queue_name: string
        """
        queue = self._get_queue(queue_name)
        queue.purge()

    def push(self, queue_name, task_id, data):
        """
        Pushes a task onto the queue.

        :param queue_name: The name of the queue. Usually handled by the
            ``Gator`` instance.
        :type queue_name: string

        :param task_id: The identifier of the task.
        :type task_id: string

        :param data: The relevant data for the task.
        :type data: string
        """
        # SQS doesn't let you specify a task id.
        queue = self._get_queue(queue_name)
        res = queue.send_message(MessageBody=data)
        return res.get("MessageId")

    def pop(self, queue_name):
        """
        Pops a task off the queue.

        :param queue_name: The name of the queue. Usually handled by the
            ``Gator`` instance.
        :type queue_name: string

        :returns: The data for the task.
        :rtype: string
        """
        queue = self._get_queue(queue_name)
        messages = queue.receive_messages(MaxNumberOfMessages=1)

        if messages:
            message = messages[0]
            data = message.body
            return data

    def get(self, queue_name, task_id):
        """
        Unsupported, as SQS does not include this functionality.
        """
        raise NotImplementedError(
            "SQS does not support fetching a specific message off the queue."
        )



import importlib

from .exceptions import UnknownModuleError, UnknownCallableError


def determine_module(func):
    """
    Given a function, returns the Python dotted path of the module it comes
    from.

    Ex::

        from random import choice
        determine_module(choice) # Returns 'random'

    :param func: The callable
    :type func: function

    :returns: Dotted path string
    """
    return func.__module__


def determine_name(func):
    """
    Given a function, returns the name of the function.

    Ex::

        from random import choice
        determine_name(choice) # Returns 'choice'

    :param func: The callable
    :type func: function

    :returns: Name string
    """
    if hasattr(func, '__name__'):
        return func.__name__
    elif hasattr(func, '__class__'):
        return func.__class__.__name__

    # This shouldn't be possible, but blow up if so.
    raise AttributeError("Provided callable '{}' has no name.".format(
        func
    ))


def import_module(module_name):
    """
    Given a dotted Python path, imports & returns the module.

    If not found, raises ``UnknownModuleError``.

    Ex::

        mod = import_module('random')

    :param module_name: The dotted Python path
    :type module_name: string

    :returns: module
    """
    try:
        return importlib.import_module(module_name)
    except ImportError as err:
        raise UnknownModuleError(str(err))


def import_attr(module_name, attr_name):
    """
    Given a dotted Python path & an attribute name, imports the module &
    returns the attribute.

    If not found, raises ``UnknownCallableError``.

    Ex::

        choice = import_attr('random', 'choice')

    :param module_name: The dotted Python path
    :type module_name: string

    :param attr_name: The attribute name
    :type attr_name: string

    :returns: attribute
    """
    module = import_module(module_name)

    try:
        return getattr(module, attr_name)
    except AttributeError as err:
        raise UnknownCallableError(str(err))
